
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import java.util.prefs.Preferences;

/**
 * This class serves as a collection of preferences for this program.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version December 9, 2023
 */
public class AppPreferences implements  jahuwaldt.swing.Preferences {

	//  The preferences for this application are stored as a properties list.
    private final Preferences prefs = Preferences.userNodeForPackage(this.getClass());
	
    //  Use a change-listener system to keep the preferences dialog updated
    private final PropertyChangeSupport propertyChange = new PropertyChangeSupport(this);
	
	//  A reference to the window containing the application preferences.
//	private PrefsDialog prefsDialog = null;
		
    //  Preference keys
    private static final String LASTPATH = "lastPath";
    
	/**
	*  Returns the file path to the parent of the last referenced file.
	*  Returns null if no last path could be found.
    * 
     * @return The file path to the parent of the last referenced file.
	*/
    @Override
	public String getLastPath() {
		return prefs.get(LASTPATH, null);
	}
	
	/**
	*  Set the last file path referenced by the user.  This is the path
	*  to the last parent of the last referenced file.
    * 
     * @param path The last file path referenced by the user.
	*/
    @Override
	public void setLastPath(String path) {
		String oldPath = getLastPath();
		if ( (nonNull(path) && !path.equals(oldPath)) || (isNull(path) && nonNull(oldPath)) ) {
			prefs.put(LASTPATH, path);
            propertyChange.firePropertyChange(LASTPATH, oldPath, path);
		}
	}

	/**
	* Return the preference with the specified key String.
	*
	*  @param key  The key String identifying the preference to be retrieved.
     * @return The preference with the specified key String.
	*/
    @Override
	public String get(String key) {
		return prefs.get(key, null);
	}
	
	/**
	* Set the preference with the specified key String.
	*
	*  @param key  The key String identifying the preference to be set.
	*  @param value THe String value to store as the preference.
	*/
    @Override
	public void set(String key, String value) {
		prefs.put(key, value);
	}
	

    public synchronized void addPropertyChangeListener(PropertyChangeListener l) {
        propertyChange.addPropertyChangeListener(l);
    }

    public synchronized void addPropertyChangeListener(String propertyName, PropertyChangeListener l) {
        propertyChange.addPropertyChangeListener(propertyName, l);
    }

    public synchronized void removePropertyChangeListener(PropertyChangeListener l) {
        propertyChange.removePropertyChangeListener(l);
    }

    public synchronized void removePropertyChangeListener(String propertyName, PropertyChangeListener l) {
        propertyChange.removePropertyChangeListener(propertyName, l);
    }

	/**
	*  Method that displays a dialog that allows the user to change the application preferences.
	*  If the application is running in batch mode, this method does nothing.
	*/
    @Override
	public void showPreferenceDialog() {
/*		if (app.getGUIApplication() != null) {
			if (prefsDialog != null)
				prefsDialog.setVisible(true);
				
			else {
				prefsDialog = new PrefsDialog();
				AppUtilities.positionWindow( prefsDialog, prefsDialog.getWidth(), prefsDialog.getHeight() );
				prefsDialog.setVisible(true);
			}
		}
*/
	}

}
