
import jahuwaldt.js.datareader.DataCase;
import jahuwaldt.js.datareader.DataSet;
import jahuwaldt.js.datareader.UnitParameter;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.FileDialog;
import java.awt.event.ActionEvent;
import java.io.File;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import static java.util.Objects.isNull;
import static java.util.Objects.nonNull;
import java.util.ResourceBundle;
import javax.measure.quantity.Dimensionless;
import javax.measure.unit.Unit;
import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.XYPlot;

/**
 * An application window that displays a quick plot of the data contained in a DataSet
 * made up of DataCase objects containing the same set of ArrayParam objects and includes
 * a list of pre-defined plots. This allows "overlay" plots of identical parameters from
 * different "runs" or cases.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version December 22, 2023
 */
public class PlotDefPlotWindow extends PlotWindow {

    private static final long serialVersionUID = 1L;

    //	The list of plot definitions from the user.
    private final PlotDefinitionList _plotDefs;
    
    //	Mapping of plot definitions to parameter database index numbers.
    private final Map<String, Integer> _varName2Index = new HashMap<>();
    
    //	The JList containing the plot definitions.
    private JList _plotDefsJList = null;

    /**
     * Creates a new instance of a data plot window using the specified list of data
     * arrays.
     *
     * @param title Title of the window.
     * @param data A DataCase containing the cases to be plotted (all cases must have the
     * same set of parameters).
     * @param plotDefs A list of plot definitions.
     * @throws java.lang.NoSuchMethodException
     * @throws IllegalArgumentException if the input data set can not be plotted.
     */
    public PlotDefPlotWindow(String title, DataCase data, PlotDefinitionList plotDefs)
            throws NoSuchMethodException, IllegalArgumentException {
        super(title, changeUnits(data, plotDefs));

        //	Assign class variables.
        _plotDefs = plotDefs;

        //  Initialize this plot window.
        initialize(data);
    }

    /**
     * Creates a new instance of a data plot window using the specified list of data
     * arrays.
     *
     * @param title Title of the window.
     * @param data DataSet containing the cases to be plotted (all cases must have the
     * same set of parameters).
     * @param plotDefs A list of plot definitions.
     * @throws java.lang.NoSuchMethodException
     * @throws IllegalArgumentException if the input data set can not be plotted.
     */
    public PlotDefPlotWindow(String title, DataSet data, PlotDefinitionList plotDefs)
            throws NoSuchMethodException, IllegalArgumentException {
        super(title, changeUnits(data,plotDefs));

        //	Assign class variables.
        _plotDefs = plotDefs;

        //  Initialize this plot window.
        initialize(data.get(0));
    }

    /**
     * Used to initialize this plot window.
     */
    private void initialize(DataCase sampleCase) throws NoSuchMethodException, IllegalArgumentException {

        if (!_plotDefs.isEmpty()) {
            //	Layout the plot window.
            layoutWindow();

            //	Map the plot definitions to index numbers in the parameter database.
            int numPlotDefs = _plotDefs.size();
            for (int i = 0; i < numPlotDefs; ++i) {
                PlotDefinition plotDef = _plotDefs.get(i);
                String name = plotDef.independentName;
                int idx = sampleCase.getIndexFromName(name);
                if (idx >= 0)
                    _varName2Index.put(name, idx);

                name = plotDef.dependentName;
                idx = sampleCase.getIndexFromName(name);
                if (idx >= 0)
                    _varName2Index.put(name, idx);
            }

            //	Set the plot to the 1st plot defintiion.
            handlePlotChange(0, false);
        }
        
    }

    private static DataSet changeUnits(DataSet data, PlotDefinitionList plotDefLst) throws IllegalArgumentException {
        //  Loop over all the cases in the data set.
        for (DataCase aCase : data) {
            changeUnits(aCase, plotDefLst);
        }
        return data;
    }

    private static DataCase changeUnits(DataCase aCase, PlotDefinitionList plotDefLst) throws IllegalArgumentException {
        
        if (!plotDefLst.isEmpty()) {
            int numPlotDefs = plotDefLst.size();

            for (int i = 0; i < numPlotDefs; ++i) {
                PlotDefinition plotDef = plotDefLst.get(i);
                
                //  Change the independent variable units.
                String name = plotDef.independentName;
                int idx = aCase.getIndexFromName(name);
                if (idx >= 0) {

                    //  Change the units if necessary.
                    Unit newUnit = plotDef.independentSrcUnit;
                    UnitParameter param = (UnitParameter) aCase.get(idx);
                    Unit oldUnit = param.getUnit();

                    //  Change the units only if the original data did not have units and
                    //  the new unit is not dimensionless.
                    if (Dimensionless.UNIT.equals(oldUnit) && !Dimensionless.UNIT.equals(newUnit))
                        aCase.set(idx, param.changeTo(newUnit));
                }
                
                //  Change the dependent variable units.
                name = plotDef.dependentName;
                idx = aCase.getIndexFromName(name);
                if (idx >= 0) {

                    //  Change the units if necessary.
                    Unit newUnit = plotDef.dependentSrcUnit;
                    UnitParameter param = (UnitParameter) aCase.get(idx);
                    Unit oldUnit = param.getUnit();

                    //  Change the units only if the original data did not have units and
                    //  the new unit is not dimensionless.
                    if (Dimensionless.UNIT.equals(oldUnit) && !Dimensionless.UNIT.equals(newUnit))
                        aCase.set(idx, param.changeTo(newUnit));
                }
            }
        }
        
        return aCase;
    }

    /**
     * Method that lays out the contents of this window.
     */
    private void layoutWindow() throws NoSuchMethodException {

        //	Create a list from the plot definitions.
        int size = _plotDefs.size();
        String[] labels = new String[size];
        for (int i = 0; i < size; ++i) {
            labels[i] = _plotDefs.get(i).getListLabel();
        }

        _plotDefsJList = new JList(labels);
        _plotDefsJList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        _plotDefsJList.setVisibleRowCount(1);
        _plotDefsJList.setLayoutOrientation(JList.HORIZONTAL_WRAP);
        _plotDefsJList.setPrototypeCellValue("<html><center>Longish Name<br>vs.<br>Longish Name</center></html>");
        _plotDefsJList.setSelectedIndex(0);
        _plotDefsJList.addListSelectionListener(new ListSelectionListener() {
            @Override
            public void valueChanged(ListSelectionEvent e) {
                if (e.getValueIsAdjusting() == false) {
                    JList list = (JList) e.getSource();
                    int idx = list.getSelectedIndex();
                    if (idx >= 0) {
                        //	Change plot to the selected parameters.
                        handlePlotChange(list.getSelectedIndex(), true);
                    }
                }
            }
        });

        //	Put the list in a scroll pane.
        JScrollPane listScroller = new JScrollPane(_plotDefsJList);
        listScroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        listScroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);

        //	Create a panel to contain the plot definition list.
        Box listPanel = Box.createHorizontalBox();
        listPanel.setAlignmentX(Box.LEFT_ALIGNMENT);
        listPanel.setBorder(BorderFactory.createEmptyBorder(4, 15, 4, 15));
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
        JLabel label = new JLabel(resBundle.getString("plotDefListLabel"), JLabel.RIGHT);
        listPanel.add(label);
        listPanel.add(listScroller);

        //	Add the list panel to the top of the center panel (so it is above the plot).
        listPanel.add(Box.createGlue());
        _centerPanel.add(listPanel, BorderLayout.NORTH);


        //	Find the export sub-menu.
        JMenuBar menuBar = this.getJMenuBar();
        JMenu fileMenu = menuBar.getMenu(0);
        JMenu exportMenu = null;
        String exportLabel = resBundle.getString("exportAsItemText");
        int numItems = fileMenu.getItemCount();
        for (int i = 0; i < numItems; ++i) {
            JMenuItem item = fileMenu.getItem(i);
            if (nonNull(item) && item.getText().equals(exportLabel)) {
                exportMenu = (JMenu) item;
                break;
            }
        }
        
        if (exportMenu == null)
            throw new NullPointerException("Couldn't find export menu item.");
        
        else {
            //	Add an item for exporting the plot definition plots to image files.
            JMenuItem menuItem = new JMenuItem(resBundle.getString("exportDefPlots2PNG"));
            menuItem.addActionListener(AppUtilities.getActionListenerForMethod(this, "handleExportDefPlots"));
            menuItem.setEnabled(true);
            exportMenu.add(menuItem);
        }
    }
    
    //	Used to hold a single index being passed to the setDependents() method.
    private final int[] _indicies = new int[1];
    private final Unit[] _units = new Unit[1];

    /**
     * Method that deals with the user changing which plot definition is being plotted.
     *
     * @param plotDefIndex The index into the list of plot definitions.
     * @param reportErrors Set to <code>true</code> to report errors to the user or
     * <code>false</code> to silently ignore errors.
     */
    private void handlePlotChange(int plotDefIndex, boolean reportErrors) {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

        //	Get the index to the independent variable.
        PlotDefinition plotDef = _plotDefs.get(plotDefIndex);
        String name = plotDef.independentName;
        Integer i = _varName2Index.get(name);
        if (isNull(i)) {
            if (reportErrors)
                JOptionPane.showMessageDialog(this,
                        MessageFormat.format(resBundle.getString("missingIndepMsg"), name),
                        resBundle.getString("missingIndepTitle"),
                        JOptionPane.ERROR_MESSAGE);
            _plotDefsJList.clearSelection();
            return;
        }
        int indepIndex = i;

        //	Get the index to the dependent variable.
        name = plotDef.dependentName;
        i = _varName2Index.get(name);
        if (isNull(i)) {
            if (reportErrors)
                JOptionPane.showMessageDialog(this,
                        MessageFormat.format(resBundle.getString("missingDepMsg"), name),
                        resBundle.getString("missingDepTitle"),
                        JOptionPane.ERROR_MESSAGE);
            _plotDefsJList.clearSelection();
            return;
        }
        _indicies[0] = i;
        _units[0] = plotDef.dependentPlotUnit;

        //	Change the plotted data.
        setIndependent(indepIndex, plotDef.independentPlotUnit);
        setDependents(_indicies, _units);
        regeneratePlot();

        //	Change the chart title.
        ChartPanel chartPanel = getChartPanel();
        JFreeChart chart = chartPanel.getChart();
        chart.setTitle(plotDef.title);

        //	Change the axis labels.
        XYPlot xyplot = chart.getXYPlot();
        ValueAxis axis = xyplot.getDomainAxis();
        String unitStr = formatUnit(plotDef.independentPlotUnit);
        axis.setLabel(plotDef.independentDesc + unitStr);
        axis = xyplot.getRangeAxis();
        unitStr = formatUnit(plotDef.dependentPlotUnit);
        axis.setLabel(plotDef.dependentDesc + unitStr);

    }

    private static String formatUnit(Unit unit) {
        String unitStr = "";
        if (!Dimensionless.UNIT.equals(unit))
            unitStr = " (" + unit + ")";
        return unitStr;
    }
    
    /**
     * Method called whenever the user changes the selection of plotted parameters. This
     * implementation clears any selection in the plot definitions list.
     */
    @Override
    protected void notifySelectionChanged() {
        if (nonNull(_plotDefsJList)) {
            _plotDefsJList.clearSelection();
            getChartPanel().getChart().setTitle((String) null);
        }
    }

    /**
     * Method that handles the user selecting "Defined Plots..." from the Export menu.
     *
     * @param event
     */
    public void handleExportDefPlots(ActionEvent event) {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

        try {
            //	Have the user select the directory to export the plots into.
            String dir = MDIApplication.getInstance().getPreferences().getLastPath();
            if (dir != null) {
                File theDir = new File(dir);
                dir = theDir.getParent();
            }
            File theDir = AppUtilities.selectDirectory(this, FileDialog.SAVE,
                    resBundle.getString("choosePlotDir"), dir, null);
            if (isNull(theDir))
                return;

            //	Start a separate thread for plotting each defined plot.
            (new SaveDefPlotsTask(this, theDir)).execute();

            //	Save off the last directory to the preferences.
            MDIApplication.getInstance().getPreferences().setLastPath(theDir.getPath());

        } catch (Exception e) {
            AppUtilities.showException(null, resBundle.getString("unexpectedTitle"),
                    resBundle.getString("unexpectedMsg"), e);
            e.printStackTrace();
        }

    }

    /**
     * A thread that selects each pre-defined plot in order and saves them each to an
     * image file.
     */
    private class SaveDefPlotsTask extends SwingWorker<Void, Integer> {
        private boolean overWriteAll = false;
        private boolean wait = false;
        private final Component parent;
        private final File dir;

        /**
         * Constructor that take the directory that the plot images files should be saved
         * into.
         */
        public SaveDefPlotsTask(Component parent, File directory) {
            this.parent = parent;
            dir = directory;
        }

        /**
         * The method to be executed in a background thread. Loop through the defined
         * plots, selecting each one in turn and then waiting until it has been saved.
         */
        @Override
        @SuppressWarnings("SleepWhileInLoop")
        protected Void doInBackground() {
            try {

                ListModel model = _plotDefsJList.getModel();
                int numPlots = model.getSize();

                for (int i = 0; i < numPlots; ++i) {
                    if (isCancelled())
                        break;

                    //	Select the specified plot.
                    //_plotDefsJList.setSelectedIndex(i);
                    handlePlotChange(i, false);

                    //	Notify the process listener which is called on the event dispatch thread.
                    wait = true;
                    publish(i);

                    //	Now wait until the newly selected plot has been saved.
                    while (wait) {
                        Thread.sleep(10L);
                    }
                }

            } catch (InterruptedException e) {
            }

            //	Re-select the originally selected plot before stopping.
            handlePlotChange(_plotDefsJList.getSelectedIndex(), false);

            return null;

        }

        /**
         * Receives data chunks from the publish method asynchronously on the Event
         * Dispatch Thread. Takes the currently displayed plot and saves it to a PNG file.
         */
        @Override
        protected void process(List<Integer> indexes) {
            try {
                int plotDefIndex = indexes.get(indexes.size() - 1);	//	Get the last one (should only be one).

                //	Create a file name for the plot image file.
                String name = _plotDefs.get(plotDefIndex).dependentName + ".png";
                File theFile = new File(dir, name);

                //	Does this file already exist?
                boolean skip = false;
                if (!overWriteAll && theFile.exists()) {
                    //	Deal with possible user options.
                    ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                    Object[] options = {resBundle.getString("replaceAllOption"), resBundle.getString("replaceOption"),
                        resBundle.getString("stopOption"), resBundle.getString("dontReplaceOption")};
                    String msg = MessageFormat.format(resBundle.getString("fileExistsMsgFmt"), name);
                    int chosen = JOptionPane.showOptionDialog(parent, msg, resBundle.getString("warningTitle"),
                            JOptionPane.DEFAULT_OPTION, JOptionPane.ERROR_MESSAGE,
                            null, options, options[0]);
                    switch (chosen) {
                        case -1:	//	User clicked on the close box for the dialog.
                        case 0:
                            //	Replace all
                            overWriteAll = true;
                            break;
                        case 2:
                            //	Stop.
                            cancel(false);
                            skip = true;
                            break;
                        case 3:
                            //	Don't replace.
                            skip = true;
                            break;
                        default:
                            //	Replace
                            break;
                    }

                }

                if (!skip) {
                    //	Write out the PNG file.
                    ChartPanel plotPanel = getChartPanel();
                    ChartUtilities.saveChartAsPNG(theFile, plotPanel.getChart(),
                            plotPanel.getWidth(), plotPanel.getHeight());
                }

            } catch (Exception e) {
                String msg = e.getMessage();
                if (msg == null) {
                    Throwable cause = e.getCause();
                    if (cause != null)
                        msg = cause.getMessage();
                }
                ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                Object[] options = {resBundle.getString("stopOption"), resBundle.getString("continueOption")};
                int chosen = JOptionPane.showOptionDialog(parent, msg, resBundle.getString("ioErrorTitle"),
                        JOptionPane.OK_CANCEL_OPTION, JOptionPane.ERROR_MESSAGE, null, options, options[1]);
                if (chosen == 0)
                    //	User canceled any further processing.
                    cancel(false);
            }

            //	We are ready for the next one, so tell the background thread to stop waiting.
            wait = false;
        }
    }
}
