/*
 *   MatV4DataReader  -- Translates DataSets containing ArrayParams to and from MatLab V4 array files.
 *
 *   Copyright (C) 2008-2014 by Joseph A. Huwaldt
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.datareader;

import jahuwaldt.swing.AppUtilities;
import jahuwaldt.tools.math.MathTools;
import java.awt.Frame;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.DoubleBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.ShortBuffer;
import java.util.List;
import javax.measure.quantity.Dimensionless;
import javolution.context.ArrayFactory;
import javolution.text.Text;
import javolution.util.FastTable;

/**
 * This class translates data between a list of DataSet data structures and a
 * stream containing MatLab V4 array file data. All the arrays must be the same
 * length (in the same case).
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 7, 2008
 * @version April 27, 2014
 */
public class MatV4DataReader implements DataReader {

	private static final boolean DEBUG = false;
	
	//	The default name given to data sets.
	private CharSequence _defaultSetName = "MatLab V4 Data";

	//  A brief description of the data read by this reader.
	private static final String DESCRIPTION = "MatLab V4 Array";
	
	//  The preferred file extension for files of this reader's type.
	public static final String EXTENSION = "mat";
	
	//	Codes used for precision types.
	private static final int DOUBLE=0, SINGLE=1, INTEGER=2, SHORT=3;
	
	/**
	*  Returns a string representation of the object.  This will return a brief description of the format
	*  read by this reader.
	*/
    @Override
	public String toString() {
		return DESCRIPTION;
	}
	
	/**
	*  Returns the preferred file extension (not including the ".") for files of
	*  this DataReader's type.
	*/
    @Override
	public String getExtension() {
		return EXTENSION;
	}
	
	/**
	* Compares this object with the specified object for order based on the
	* <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
	* zero, or a positive integer as this object is less than, equal to, or
	* greater than the specified object.
	*/
    @Override
	public int compareTo(DataReader o) {
		return this.toString().compareTo(o.toString());
	}
	
	/**
	*  Method that determines if this reader can read data from the specified input stream.
	*
	*  @param  pathName The path to the file to be read.
	*  @param  input  An input stream containing the data to be read.  Any methods that
	*                 read from this stream must first set a mark and then reset back to
	*                 that mark before the method returns (even if it returns with an
	*                 exception).
	*  @returns DataReader.NO if the file is not recognized or DataReader.MAYBE if the
	*           appears to be readable.
	*/
    @Override
	public int canReadData(String pathName, BufferedInputStream input) throws IOException {

		int response = MAYBE;
		try {
			//	Check the format of the file.  If an exception is thrown, then
			//	we can't read the file.
			try {
				//	Mark the current place in the input stream so we can return to it when we are done.
				input.mark(5*4);	//	We will be reading in a 20 byte header (five 4-byte integers).
			
				//	Read in the header data, checking for validity.
				readHeaderData(input);
			
			} finally {
				//	Reset the buffer back to where it was before reading from it.
				input.reset();
			}
			
		} catch (Exception e) {
			response = NO;
		}
		
		return response;
	}
	
	/**
	*  Returns true.  This class can write data to a MatLab V4 formatted file stream.
	*/
    @Override
	public boolean canWriteData() {
		return true;
	}
	

	/**
	*  The MatLab V4 format can save only a single case of arrays of all the same length.
	*
	*  @param parent  Determines the Frame in  which the dialog is displayed; if null,
	*                 or if the parentComponent has no Frame, a default Frame is used.
	*  @param data    The input data set that is to be selected for saving.
	*  @return  A list of DataSet objects containing the selected data to be saved.
	*           Could return null if the user selects nothing.
	*/
    @Override
	public List<DataSet> selectDataForSaving(Frame parent, List<DataSet> data) {
	
		if (data == null || data.size() < 1)	return null;

		DataSet selected;
		
		if (data.size() == 1 && data.get(0).size() == 1)
			//  If there is just one case, save that.
			selected = data.get(0);
		
		else {
			//  If there is more than one case, 
			//	ask the user to select cases to export.
			SelectCasesDialog dialog = new SelectCasesDialog(parent, "Select a case to Export",
												"Select a single case from the list below:", data, true, true);
			dialog.setLocation(AppUtilities.dialogPosition(dialog));
			dialog.setVisible(true);

			//	Retrieve the selected case.
			selected = (DataSet)dialog.getSelected();
			dialog.dispose();
		}
		if (selected == null || selected.size() < 1) return null;

		FastTable<DataSet> output = FastTable.newInstance();
		output.add(selected);

		return output;
	}
	
	
	/**
	*  Method that sets the default set name to use for
	*  MatLab V4 array data sets.
	*/
    @Override
	public void setDefaultSetName(CharSequence name) {
		_defaultSetName = name;
	}

	
	/**
	*  Method that reads in MatLab V4 array data
	*  from the specified input stream and returns that data as a list
	*  of {@link DataSet} objects (containing a single {@link DataCase} containing an
	*  {@link ArrayParam} for each column in the MatLab array).
	*
	*  @param  pathName The path to the file to be read.
	*  @param   input  An input stream containing the MatLab V4 array data.
	*  @returns A list of {@link DataSet} objects that contains the data read
	*           in from the specified stream (will contain a single data set
	*           which will contain a single case containing ArrayParams for
	*           each column in the MatLab 2D array).
	*  @throws IOException If there is a problem reading the specified stream.
	*/
    @Override
	public List<DataSet> read(String pathName, InputStream input) throws IOException {

		//	Create an empty list to store DataSets into.
		FastTable<DataSet> dataSets = FastTable.newInstance();

		//	Read in the header data.
		HeaderData hd = readHeaderData(input);
		
		//	Read in the array name.
		Text name = readArrayName(hd, input);
		
		//	Create a list of empty arrays that have the right sizes.
		FastTable<double[]> arrs = FastTable.newInstance();
		for (int i=0; i < hd.mrows; ++i) {
			arrs.add(new double[hd.ncols]);
		}
		
		//	Read in the 2D array from the MatLab file.
		readArray(hd, arrs, input);
		
        //  Convert from simple arrays to ArrayParam instances.
        FastTable<ArrayParam> arrParams = FastTable.newInstance();
		for (int i=0; i < hd.mrows; ++i) {
			ArrayParam array = ArrayParam.valueOf("", Dimensionless.UNIT, arrs.get(i));
			arrParams.add(array); 
		}
        
		//	MatLab V4 files contain only a single case, so create that.
		DataCase aCase = DataCase.newInstance(name);
		aCase.addAll(arrParams);
		FastTable.recycle(arrs);
        FastTable.recycle(arrParams);
		
		//	MatLab V4 files contain only a single data set, so create that.
		DataSet set = DataSet.newInstance(_defaultSetName);

		//	Add the case to the set.
		set.add(aCase);

		//	Add the set to the list of data sets being output.
		dataSets.add(set);

		return dataSets;
	}

	/**
	*  Method that writes out the data stored in the specified
	*  {@link DataSet} object to the specified output stream in MatLab V4 array format.
	*  Only the 1st case in the 1st data set in the specified list is written out since
	*  the MatLab V4 array format does not support multiple cases.  Only arrays are
	*  written out since the MatLab V4 array format does not support scalars or
	*  text notes.  All arrays must be the same length.
	*
	*  @param  output  The output stream to which the data is to be written.
	*  @param  data    A list of {@link DataSet} objects containing data to be written out.
	*                  Only the 1st case in the 1st data set in the list is written out.
	*  @throws IOException If there is a problem writing to the specified stream.
	*/
    @Override
	public void write( OutputStream output, List<DataSet> data ) throws IOException {
	
		if (data.size() < 1)	return;		//	No data to be written out.

		//	Get the 1st case in the 1st data set.  That's all the MatLab file can hold.
		DataSet set = data.get(0);
		DataCase aCase = set.get(0);
		
		//	Get a list of all the ArrayParam objects (since that's all the MatLab file can hold).
		FastTable<ArrayParam> arrays = (FastTable<ArrayParam>)aCase.getAllArrays();
		if (arrays.size() < 1)	return;		//	No array data to be written out.
		
		//	Create the header data.
		int mrows = arrays.size();
		int ncols = arrays.get(0).size();
		Text name = Text.valueOf(aCase.getName());
		
		//	Make sure there are no illegal characters in the name.
		name = name.replace(".","_");
		
		HeaderData hd = new HeaderData();
		hd.byteOrder = ByteOrder.BIG_ENDIAN;	//	Java is always big-endian.
		hd.precision = DOUBLE;
		hd.mrows = mrows;
		hd.ncols = ncols;
		hd.namelen = name.length()+1;
		
		//	Write out the header.
		writeHeaderData(output, hd);
		
		//	Write the array title.
		writeArrayName(output, name);
		
		//	Write out the array of data.
		writeArray(output, arrays);
		
	}


	/**
	*  Write out the 20 byte header data to the MatLab V4 output stream.  The header consists
	*  of five 4-byte integers that describe the contents of the file.
	*
	*  @param out  Output stream to the MatLab V4 file.
	*  @param hd   The header data to write out.
	*/
	private void writeHeaderData(OutputStream out, HeaderData hd) throws IOException {
	
		//	Create a byte buffer.
		int size = 5*4;
		byte[] bytes = ArrayFactory.BYTES_FACTORY.array(size);
		try {
			ByteBuffer buf = ByteBuffer.wrap(bytes);
			
			int type = 0;
			if (hd.byteOrder == ByteOrder.BIG_ENDIAN)	type += 1000;
			//	Data is always column-wise (type += 100*row-wise).
			type += 10*hd.precision;
			//	Data is always numeric (type += textual).
			
			//	Fill in the header data.
			buf.putInt(type);
			buf.putInt(hd.mrows);
			buf.putInt(hd.ncols);
			buf.putInt(0);				//	Only real data (no imaginary values).
			buf.putInt(hd.namelen);
			
			out.write(bytes, 0, size);
		
		} finally {
			ArrayFactory.BYTES_FACTORY.recycle(bytes);
		}
	}
	
	/**
	*  Write out the name of the array.
	*
	*  @param out  Output stream to the MatLab V4 file.
	*  @param name The name of the array to be written out.
	*/
	private void writeArrayName(OutputStream output, Text name) throws IOException {
	
		// Wrap the output stream in a Writer that converts the String to bytes.
		OutputStreamWriter writer = new OutputStreamWriter(output);
		
		//	Write out the name.
		name.print(writer);
		writer.flush();
		
		//	Write out a terminating byte.
		output.write(0);
		
	}
	
	/**
	*  Write out the array of data to the output stream.
	*
	*  @param out  Output stream to the MatLab V4 file.
	*  @param arr  The array to be written out.
	*/
	private void writeArray(OutputStream output, List<ArrayParam> arrays) throws IOException {
		
		//	Get the number of rows and columns.
		ArrayParam param = arrays.get(0);
		int mrows = arrays.size();
		int ncols = param.size();
		
		//	Allocate a byte buffer for converting doubles to bytes.
		int size = mrows*8;
		byte[] bytes = ArrayFactory.BYTES_FACTORY.array(size);
		ByteBuffer byteBuf = ByteBuffer.wrap(bytes);
		
		//	View the byte buffer as a double buffer for convenience.
		DoubleBuffer buf = byteBuf.asDoubleBuffer();
		
		//	Loop over all the columns and write out the data, one column at a time.
		double[] arr = ArrayFactory.DOUBLES_FACTORY.array(mrows);
		try {
		
			for (int i=0; i < ncols; ++i) {
			
				//	Transpose the data as we go.
				for (int j=0; j < mrows; ++j) {
					param = arrays.get(j);
					arr[j] = param.getValue(i);
				}
				
				//	Write out the data.
				buf.rewind();
				buf.put(arr, 0, mrows);
				output.write(bytes, 0, size);
			}
		
		} finally {
			ArrayFactory.DOUBLES_FACTORY.recycle(arr);
			ArrayFactory.BYTES_FACTORY.recycle(bytes);
		}
	}

	
	/**
	*  Read in the 20 byte header data and check for valid numbers.  The header consists
	*  of five 4-byte integers that describe the contents of the file.
	*
	*  @param is  MatLab V4 binary formatted input stream.
	*  @return The header data extracted from the input stream.
	*/
	private HeaderData readHeaderData(InputStream is) throws IOException {
	
		//	Create a header object.
		HeaderData hd = new HeaderData();
		
		//	Read in the 20 byte header.
		int size = 5*4;
		byte[] bytes = ArrayFactory.BYTES_FACTORY.array(size);
		int numRead = is.read(bytes, 0, size);
		if (numRead != size)	throw new IOException("Could not read the MatLab V4 file header.");
		
		//	Determine the byte order.
		if (bytes[2] == 0)
			hd.byteOrder = ByteOrder.LITTLE_ENDIAN;
		else
			hd.byteOrder = ByteOrder.BIG_ENDIAN;
		
		try {
			//	Create a byte buffer to read and translate the bytes.
			ByteBuffer buffer = ByteBuffer.wrap(bytes);
			buffer = buffer.order(hd.byteOrder);			//	The buffer will convert byte order if necessary.
		
			//	Read in the type codes.
			int type = buffer.getInt();
			if (hd.byteOrder == ByteOrder.BIG_ENDIAN) {
				if (type < 1000)
					throw new IOException("Unrecognized MatLab V4 type code.");
				type -= 1000;
			}
			
			//	Is the data column-wise or row-wise.
			int columnWise = (int)MathTools.roundToPlace(type, 2);
			if (columnWise != 0)
				throw new IOException("Can only read column-wise data from MatLab V4 file.");
			
			//	What form is the data in (double, single, various integer types).
			hd.precision = (int)MathTools.roundToPlace(type, 1);
			if (hd.precision < 0 || hd.precision > 3)
				throw new IOException("Unrecognized precision value in MatLab V4 file.  Must be 0, 1, 2, or 3.");
			type -= 10*hd.precision;
			
			//	Is the data numeric or textual?
			if (type != 0)
				throw new IOException("Can only read numeric data, not textual from MatLab V4 file.");
			
			//	Read in the number of rows.
			hd.mrows = buffer.getInt();
			if (hd.mrows <= 0)	throw new IOException("Bad number of rows in MatLab V4 file. Must be at least 1.");
			
			//	Read in the number of columns.
			hd.ncols = buffer.getInt();
			if (hd.ncols <= 0)	throw new IOException("Bad number of columns in MatLab V4 file. Must have at least 1.");
			
			//	Read in the imaginary number flag.
			int imagf = buffer.getInt();
			if (imagf != 0)	throw new IOException("Can not read imaginary numbers from MatLab V4 file.");
			
			hd.namelen = buffer.getInt();
			if (hd.namelen < 2)	throw new IOException("Bad name length in MatLab V4 file.");
			
		} catch (BufferUnderflowException e) {
			//	Shouldn't be possible.
			e.printStackTrace();
			
		} finally {
			ArrayFactory.BYTES_FACTORY.recycle(bytes);
		}
		
		if (DEBUG) {
			System.out.println("hd.byteOrder = " + hd.byteOrder);
			System.out.println("hd.precision = " + hd.precision);
			System.out.println("hd.mrows = " + hd.mrows);
			System.out.println("hd.ncols = " + hd.ncols);
			System.out.println("hd.namelen = " + hd.namelen);
		}
		
		return hd;
	}
	
	
	/**
	*  Method that reads in the name of the array from the MatLab V4 file and
	*  returns it as a String.
	*
	*  @param hd  The header data extracted from the input stream.
	*  @param is  MatLab V4 binary formatted input stream.
	*  @return A String containing the name of the array in the file.
	*/
	private Text readArrayName(HeaderData hd, InputStream is) throws IOException {
		String name = null;
		
		byte[] bytes = ArrayFactory.BYTES_FACTORY.array(hd.namelen);
		java.util.Arrays.fill(bytes, hd.namelen, bytes.length, (byte)0);
			
		try {
		
			//	Read in the array name bytes.
			int num = is.read(bytes, 0, hd.namelen);
			if (num != hd.namelen)	throw new IOException("Could not read array name from MatLab V4 file.");
			
			//	Convert the bytes to a String.
			BufferedReader reader = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(bytes)));
			name = reader.readLine();
			if (name == null)	throw new IOException("Trouble reading the array name from the MatLab V4 file.");
			
			if (DEBUG) {
				System.out.println(" name = " + name.trim());
			}
			
		} finally {
			ArrayFactory.BYTES_FACTORY.recycle(bytes);
		}
		
		return Text.valueOf(name).trim();
	}
	
	/**
	*  Method that reads in the array of numbers from the MatLab V4 file and
	*  returns a 2D array of double values.
	*
	*  @param hd   The header data extracted from the input stream.
	*  @param arrs A list of arrays already sized to hold the data.
	*  @param is   MatLab V4 binary formatted input stream.
	*  @return A 2D array of data from the the file.
	*/
	private void readArray(HeaderData hd, List<double[]> arrs, InputStream is) throws IOException {
	
		//	Determine the size (in bytes) of the array in the input stream.
		int wordSize = 4;		//	Number of bytes per data value in the stream.
		switch(hd.precision) {
			case DOUBLE:	wordSize = 8;	break;
			
			case SINGLE:
			case INTEGER:	wordSize = 4;	break;
			
			case SHORT:		wordSize = 2;	break;
		}
		int size = wordSize*hd.ncols*hd.mrows;
		
		if (DEBUG) {
			System.out.println(" wordSize = " + wordSize);
			System.out.println(" size = " + size);
		}
		
		//	Read in the remaining data into a byte buffer.
		byte[] bytes = ArrayFactory.BYTES_FACTORY.array(size);
		try {
			int num = is.read(bytes, 0, size);
			if (num != size)	throw new IOException("Unexpected end of file encountered in MatLab V4 file.");
			
			//	Create a byte buffer to translate the bytes.
			ByteBuffer buffer = ByteBuffer.wrap(bytes);
			buffer = buffer.order(hd.byteOrder);			//	The buffer will convert byte order if necessary.
			
			//	Translate all the values from the byte buffer to an array of doubles.
			switch(hd.precision) {
				case DOUBLE:
					translateBytes(hd, arrs, buffer.asDoubleBuffer());
					break;
					
				case SINGLE:
					translateBytes(hd, arrs, buffer.asFloatBuffer());
					break;
					
				case INTEGER:
					translateBytes(hd, arrs, buffer.asIntBuffer());
					break;
					
				case SHORT:
					translateBytes(hd, arrs, buffer.asShortBuffer());
					break;
			}
		
		} finally {
			ArrayFactory.BYTES_FACTORY.recycle(bytes);
		}
		
	}
	
	
	/**
	*  Translate a ByteBuffer of doubles into an array of doubles.
	*/
	private void translateBytes(HeaderData hd, List<double[]> arrs, DoubleBuffer buffer) {
		
		//	Allocate memory for the double array.
		int mrows = hd.mrows;
		int ncols = hd.ncols;
		
		//	Translate the data from the buffer, transposing as we go.
		for (int i=0; i < ncols; ++i) {
			for (int j=0; j < mrows; ++j) {
				arrs.get(j)[i] = buffer.get();
			}
		}
		
	}
	
	/**
	*  Translate a ByteBuffer of floats into an array of doubles.
	*/
	private void translateBytes(HeaderData hd, List<double[]> arrs, FloatBuffer buffer) {
		
		//	Allocate memory for the double array.
		int mrows = hd.mrows;
		int ncols = hd.ncols;
		
		//	Translate the data from the buffer, transposing as we go.
		for (int i=0; i < ncols; ++i) {
			for (int j=0; j < mrows; ++j) {
				arrs.get(j)[i] = buffer.get();
			}
		}
		
	}
	
	/**
	*  Translate a ByteBuffer of integers into an array of doubles.
	*/
	private void translateBytes(HeaderData hd, List<double[]> arrs, IntBuffer buffer) {
		
		//	Allocate memory for the double array.
		int mrows = hd.mrows;
		int ncols = hd.ncols;
		
		//	Translate the data from the buffer, transposing as we go.
		for (int i=0; i < ncols; ++i) {
			for (int j=0; j < mrows; ++j) {
				arrs.get(j)[i] = buffer.get();
			}
		}
		
	}
	
	/**
	*  Translate a ByteBuffer of shorts into an array of doubles.
	*/
	private void translateBytes(HeaderData hd, List<double[]> arrs, ShortBuffer buffer) {
		
		//	Allocate memory for the double array.
		int mrows = hd.mrows;
		int ncols = hd.ncols;
		
		//	Translate the data from the buffer, transposing as we go.
		for (int i=0; i < ncols; ++i) {
			for (int j=0; j < mrows; ++j) {
				arrs.get(j)[i] = buffer.get();
			}
		}
		
	}
	
	
	/**
	*  Class that contains the header data for the MatLab V4 file.
	*/
	private class HeaderData {
		//	The byte order of the input stream.
		ByteOrder byteOrder;
		
		//	Precision of the data in the file (0=double, 1=single, 2 = int, 3 = short).
		int precision = DOUBLE;
		
		//	Number of rows in the array (this program interprets these as one 1D array parameter per row).
		int mrows;
		
		//	Number of columns in the array (this program interprets this as the elements in a 1D array).
		int ncols;
		
		//	Length of the array name + 1.
		int namelen;
	}

}



