/*
 *   OTISDataReader  -- Reads delimited columns of data from an OTIS plot file.
 *
 *   Copyright (C) 2011-2014 by Joseph A. Huwaldt
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.datareader;

import jahuwaldt.js.util.TextTokenizer;
import jahuwaldt.swing.AppUtilities;
import java.awt.Frame;
import java.io.*;
import java.util.List;
import java.util.NoSuchElementException;
import javax.measure.quantity.Dimensionless;
import javolution.text.Text;
import javolution.text.TypeFormat;
import javolution.util.FastTable;


/**
 * This class translates data between a list of DataSet data structures and a stream
 * containing delimited columns of array data from OTIS.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version April 27, 2014
 */
public class OTISDataReader implements DataReader {

	//	The default name given to DAT data sets.
	private CharSequence _defaultSetName = "OTIS Data";

	//	The delimiters to use to separate columns in the file.
	//	Use space, tab, and comma as delimiters.
	private static final String DELIMITERS = " \t,";
	
	//  A brief description of the data read by this reader.
	private static final String DESCRIPTION = "OTIS Data";
	
	//  The preferred file extension for files of this reader's type.
	public static final String EXTENSION = "op1";
	

	/**
	*  Returns a string representation of the object.  This will return a brief description of the format
	*  read by this reader.
	*/
    @Override
	public String toString() {
		return DESCRIPTION;
	}
	
	/**
	*  Returns the preferred file extension (not including the ".") for files of
	*  this DataReader's type.
	*/
    @Override
	public String getExtension() {
		return EXTENSION;
	}
	
	/**
	* Compares this object with the specified object for order based on the
	* <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
	* zero, or a positive integer as this object is less than, equal to, or
	* greater than the specified object.
	*/
    @Override
	public int compareTo(DataReader o) {
		return this.toString().compareTo(o.toString());
	}
	
	/**
	*  Method that determines if this reader can read data from the specified input stream.
	*
	*  @param  pathName The path to the file to be read.
	*  @param  input  An input stream containing the data to be read.  Any methods that
	*                 read from this stream must first set a mark and then reset back to
	*                 that mark before the method returns (even if it returns with an
	*                 exception).
	*  @returns DataReader.NO if the file is not recognized at all or DataReader.MAYBE if
	*           the file has an appropriate extension.
	*/
    @Override
	public int canReadData(String pathName, BufferedInputStream input) throws IOException {

		pathName = pathName.trim().toLowerCase();
		
		int response = NO;
		if (pathName.endsWith(".op1"))
			response = MAYBE;
		
		return response;
	}
	
	/**
	*  Returns false.  This class can not write data to an OTIS file.
	*/
    @Override
	public boolean canWriteData() {
		return false;
	}
	

	/**
	*  The OTIS format can save only a single case of arrays.
	*
	*  @param parent  Determines the Frame in  which the dialog is displayed; if null,
	*                 or if the parentComponent has no  Frame, a default Frame is used.
	*  @param data    The input data set that is to be selected for saving.
	*  @return  A list of DataSet objects containing the selected data to be saved.
	*           Could return null if the user selects nothing.
	*/
    @Override
	public List<DataSet> selectDataForSaving(Frame parent, List<DataSet> data) {
	
		if (data == null || data.size() < 1)	return null;

		DataSet selected;
		
		if (data.size() == 1 && data.get(0).size() == 1)
			//  If there is just one case, save that.
			selected = data.get(0);
		
		else {
			//  If there is more than one case, 
			//	ask the user to select cases to export.
			SelectCasesDialog dialog = new SelectCasesDialog(parent, "Select a case to Export",
												"Select a single case from the list below:", data, true, true);
			dialog.setLocation(AppUtilities.dialogPosition(dialog));
			dialog.setVisible(true);

			//	Retrieve the selected case.
			selected = (DataSet)dialog.getSelected();
			dialog.dispose();
		}
		if (selected == null || selected.size() < 1) return null;

		FastTable<DataSet> output = FastTable.newInstance();
		output.add(selected);

		return output;
	}
	
	
	/**
	*  Method that reads in OTIS delimited array data
	*  from the specified input stream and returns that data as a list
	*  of {@link DataSet} objects.
	*
	*  @param  pathName The path to the file to be read.
	*  @param   input  An input stream containing the OTIS delimited array data.
	*  @returns A list of DataSet objects that contains the data read
	*           in from the specified stream (will contain a single data set
	*           which will contain either 1 or 2 cases with ArrayParam objects for each column).
	*  @throws IOException If there is a problem reading the specified stream.
	*/
    @Override
	public List<DataSet> read(String pathName, InputStream input) throws IOException {

		//	Wrap the input stream in a line number reader.
		LineNumberReader reader = new LineNumberReader(new InputStreamReader(input));

		//	Create an empty list to store DataSets into.
		FastTable<DataSet> dataSets = FastTable.newInstance();

		//	Tab files contain only a single data set, so create that.
		DataSet set = DataSet.newInstance(_defaultSetName);

		//	Read until a non-blank line is encountered.
		String aLine = reader.readLine();
		while (aLine != null && aLine.trim().equals("")) {
			aLine = reader.readLine();
		}
		if (aLine == null)	throw new IOException("Unexpected EOF encountered on line #" + reader.getLineNumber() + ".");
		
		//	Parse out the parameter names from the 1st line.
		FastTable<Text> names = inputParameterNames(aLine);

		//	Skip the 2nd line (it is blank).
		aLine = reader.readLine();
		if (aLine == null)	throw new IOException("Unexpected EOF encountered on line #" + reader.getLineNumber() + ".");
		
		//	Begin reading data.
		for (int i=0; i < 2; ++i) {
			
			//	Create a new case.
			DataCase aCase = DataCase.newInstance("Run " + (i+1) + ": " + _defaultSetName);

			//	Now read in an array of data.
			inputDataArrays(reader, aCase, names);

			//	Add the case to the set.
			set.add(aCase);
			
			//	Is there any more data?
			reader.mark(10240);
			aLine = reader.readLine();
			while (aLine != null && aLine.trim().equals("")) {
				aLine = reader.readLine();
			}
			reader.reset();
			if (aLine == null)	break;
		}
		
		//	If there is only one run, then the "Run #" part can be left out.
		if (set.size() == 1)
			set.get(0).setName(_defaultSetName);
			
		//	Add the set to the list of data sets being output.
		dataSets.add(set);
		
		//	Clean up before leaving.
		FastTable.recycle(names);
		
		return dataSets;
	}

	/**
	*  Method that writes out all the data stored in the specified list of
	*  {@link DataSet} objects to the specified output stream in OTIS delimited format.
	*  WARNING:  This method always throws an exception.
	*
	*  @param  output  The output stream to which the data is to be written.
	*  @param  data    A list of {@link DataSet} objects containing data to be written out.
	*                  Only the 1st case in the 1st data set in the list is written out.
	*  @throws IOException If there is a problem writing to the specified stream.
	*/
    @Override
	public void write( OutputStream output, List<DataSet> data ) throws IOException {
		throw new IOException("This method has not yet been implemented");
	}


	/**
	*  Sets the default set name to use.
	*/
    @Override
	public void setDefaultSetName(CharSequence name) {
		_defaultSetName = name;
	}

	

	/**
	*  Method that reads in and parses out the parameter names from
	*  the 1st line of the file.
	*/
	private FastTable<Text> inputParameterNames(String line) throws IOException {
	
		//	Create a list of accumulate parameter names.
		FastTable<Text> list = FastTable.newInstance();
		
		//	Create a tokenizer.
		TextTokenizer tokenizer = TextTokenizer.valueOf(line, DELIMITERS);

		//	Add each label found to the list of parameter names.
		for (Text token : tokenizer)
			list.add(token);
		
		TextTokenizer.recycle(tokenizer);
		
		int length = list.size();
		if (length < 1)
			throw new IOException("No parameter names found on 1st line.");

		return list;
	}
	

	/**
	*  Method that reads in arrays of delimited data and creates array parameters
	*  out of them.
	*
	*  @param  reader  The reader for our input stream.
	*  @param  aCase   The data case that parameters should be added to.
	*  @param  names   A list of names, one for each parameter (column).
	*/
	private void inputDataArrays(LineNumberReader reader, DataCase aCase, List<Text> names) throws IOException {

		int numParams = names.size();
		int lineNum = reader.getLineNumber();

		//	Read in all the lines of the file, storing them in memory, to count how
		//	many lines (array elements) there are.
		FastTable<Text> data = FastTable.newInstance();
		String aLine;
		do {
			//	Read a line from the file.
			aLine = reader.readLine();

			//	Stop when a blank line is encountered
			if (aLine == null || aLine.startsWith("--") || aLine.trim().equals(""))
				break;

			//	Add the line to our list of lines.
			data.add(Text.valueOf(aLine));

		} while (true);

		//	Create a list of properly sized arrays, one for each parameter.
		FastTable<double[]> arrays = FastTable.newInstance();
		int numElements = data.size();
		for (int i=0; i < numParams; ++i) {
			arrays.add(new double[numElements]); 
		}

		//	Now go through the list of data strings and parse out the numbers.
		TextTokenizer tokenizer = TextTokenizer.newInstance();
		tokenizer.setDelimiters(DELIMITERS);
		int column = 0;
		try {
			for (int row=0; row < numElements; ++row, ++lineNum) {
				Text text = data.get(row);
				
				//	Create a tokenizer.
				tokenizer.setText(text);
				
				//	Loop through the line.
				for (column=0; column < numParams; ++column) {
					//	Parse out the value.
					Text token = tokenizer.nextToken();
					double value = TypeFormat.parseDouble(token.trim());

					//	Store the value.
					arrays.get(column)[row] = value;
				}
			}

		} catch (NumberFormatException e) {
			++column;
			++lineNum;
			throw new IOException("Expected a number, didn't get it: line#" + lineNum
										 + ", column#" + column + ".");

		} catch (NoSuchElementException e) {
			++lineNum;
			throw new IOException("Unexpected end-of-line on line #" + lineNum + ".");
		}
		
        //  Convert from simple arrays to ArrayParam instances.
        FastTable<ArrayParam> arrParams = FastTable.newInstance();
		for (int i=0; i < numParams; ++i) {
			ArrayParam array = ArrayParam.valueOf(names.get(i), Dimensionless.UNIT, arrays.get(i));
			arrParams.add(array); 
		}
		
		//	The list of arrays should now be filled with numbers.
		//	Put the arrays in the data case.
		aCase.addAll(arrParams);


		//	Clean up before leaving.
		TextTokenizer.recycle(tokenizer);
		FastTable.recycle(data);
		FastTable.recycle(arrays);
        FastTable.recycle(arrParams);
	}

	
}



