
/**
*   Please feel free to use any fragment of the code in this file that you need
*   in your own work. As far as I am concerned, it's in the public domain. No
*   permission is necessary or required. Credit is always appreciated if you
*   use a large chunk or base a significant product on one of my examples,
*   but that's not required either.
*
*   This code is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
*      --- Joseph A. Huwaldt
**/
package jahuwaldt.swing;

import javax.swing.JRadioButton;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.CannotRedoException;


/**
*  Class used to record the changes made to a JButtonGroup
*  object which holds a list of JRadioButtons.
*  This class is used to support an undo/redo functionality.
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author  Joseph A. Huwaldt   Date: June 6, 2008
*  @version September 16, 2012
**/
@SuppressWarnings("serial")
public class JButtonGroupEdit extends AbstractUndoableEdit {

	private JButtonGroup btnGroup;
	private JRadioButton origValue, newValue;
	private String presName;

	/**
	*  Construct an edit that records changes made to a JButtonGroup.  The
	*  originally selected button is the the button returned by the
	*  "getPreviousSelected()" method of JButtonGroup.
	*  The newly selected button is the button returned by
	*  the "getSelected()" method of JButtonGroup. 
	*
	*  @param btnGroup          The radio button group being edited.
	*  @param presentationName  Brief, human readable, description of the edit.
	**/
	public JButtonGroupEdit(JButtonGroup btnGroup, String presentationName) {
		this.btnGroup = btnGroup;
		this.origValue = (JRadioButton)btnGroup.getPreviousSelected();
		this.newValue = (JRadioButton)btnGroup.getSelected();
		this.presName = presentationName;
	}
	
	/**
	*  Return the original button selected in the button group before the edit.
	**/
	public JRadioButton originalValue() {
		return origValue;
	}
	
	/**
	*  Return the new button selected in the button group after the edit.
	**/
	public JRadioButton newValue() {
		return newValue;
	}
	
	/**
	*  Undo the edit that was made.
	**/
    @Override
	public void undo() throws CannotUndoException {
		super.undo();
		btnGroup.setSelected(originalValue(), true);
	}
	
	/**
	* Re-apply the edit, assuming that it has been undone.
	**/
    @Override
	public void redo() throws CannotRedoException {
		super.redo();
		btnGroup.setSelected(newValue(), true);
	}
	
	/**
	*  Provides a localized, human readable description of this edit suitable for use in, say, a change log.
	**/
    @Override
	public String getPresentationName() {
		return presName;
	}
}
